/*
	(C) 1995-2001 AROS - The Amiga Research OS
	$Id: rootclass.c,v 1.8 2016/09/21 03:21:28 jacadcaps Exp $
 
	Desc: Implementation of ROOTCLASS
	Lang: english
*/

#define AROS_ALMOST_COMPATIBLE

#include <exec/lists.h>
#include <exec/memory.h>
#include <proto/exec.h>
#include <proto/alib.h>
//#include <proto/arossupport.h>
#ifndef INTUITION_CLASSES_H
#   include <intuition/classes.h>
#endif
#ifndef UTILITY_HOOKS_H
#   include <utility/hooks.h>
#endif
#include <utility/utility.h>
#include <aros/asmcall.h>
#include "intuition_intern.h"
#include "morphos/boopsiextra.h"

#define IntuitionBase	(GetPrivIBase(cl->cl_UserData))

/*****i************************************************************************
 
	NAME */
AROS_UFH3(IPTR, rootDispatcher,

		  /*  SYNOPSIS */
		  AROS_UFHA(Class  *, cl,  A0),
		  AROS_UFHA(Object *, o,   A2),
		  AROS_UFHA(Msg,      msg, A1))

/*  FUNCTION
	internal !
 
	Processes all messages sent to the RootClass. Unknown messages are
	silently ignored.
 
	INPUTS
	cl - Pointer to the RootClass
	o - This object was the destination for the message in the first
		place
	msg - This is the message.
 
	RESULT
	Processes the message. The meaning of the result depends on the
	type of the message.
 
	NOTES
	This is a good place to debug BOOPSI objects since every message
	should eventually show up here.
 
	EXAMPLE
 
	BUGS
 
	SEE ALSO
 
	HISTORY
	14.09.93    ada created
 
******************************************************************************/
{
	AROS_USERFUNC_INIT

	IPTR  retval = 0;
	Class *objcl;

	switch (msg->MethodID)
	{
	case OM_NEW:
		objcl = (Class *)o;

		/* Get memory. The objects shows how much is needed.
		   (The object is not an object, it is a class pointer!) */
		if (objcl->cl_Reserved)
		{
			o = (Object *) AllocPooled((APTR)objcl->cl_Reserved,objcl->cl_InstOffset +
		                         objcl->cl_InstSize +
								 sizeof (struct _Object) +
								 sizeof (struct BOOPSIExtra));
		} else {
			o = (Object *) AllocMem(objcl->cl_InstOffset +
									 objcl->cl_InstSize +
									 sizeof (struct _Object) + sizeof (struct BOOPSIExtra),
									 MEMF_CLEAR);
		}

		if (o)
		{
			Forbid();
			++objcl->cl_ObjectCount;
			Permit();
			((struct BOOPSIExtra *)o)->UseCount = 1;
			o = (Object *)BASEROOTOBJECT(o);
			((struct _Object *)o)->o_Class = objcl;
			retval = (IPTR) BASEOBJECT(o);
		}
		break;

	case OM_DISPOSE:
		/* Free memory. Caller is responsible that everything else
		   is already cleared! */
		objcl = OCLASS(o);
		Forbid();
		if (1 == BOOPSIEXTRA(o)->UseCount)
		{
			--objcl->cl_ObjectCount;
			Permit();
			if (objcl->cl_Reserved)
			{
				FreePooled ((APTR)objcl->cl_Reserved,BOOPSIEXTRA(o),
						 objcl->cl_InstOffset + objcl->cl_InstSize + sizeof (struct _Object) + sizeof (struct BOOPSIExtra));
			} else {
				FreeMem(BOOPSIEXTRA(o),objcl->cl_InstOffset + objcl->cl_InstSize + sizeof (struct _Object) + sizeof (struct BOOPSIExtra));
			}
		}
		else
		{
			Permit();
		}
		break;

	case OM_RETAIN:
		Forbid();
		BOOPSIEXTRA(o)->UseCount++;
		Permit();
		retval = (IPTR)o;
		break;
		
	case OM_RELEASE:
		Forbid();
		if (1 == BOOPSIEXTRA(o)->UseCount)
		{
			Permit();
			DisposeObject(o);
			retval = 0;
		}
		else
		{
			BOOPSIEXTRA(o)->UseCount--;
			Permit();
			retval = (IPTR)o;
		}
		break;

	case OM_RETAINCNT:
		Forbid();
		retval = BOOPSIEXTRA(o)->UseCount;
		Permit();
		break;

	case OM_ADDTAIL:
		/* Add <o> to list. */
		AddTail (((struct opAddTail *)msg)->opat_List, (struct Node *) _OBJECT(o));
		retval = TRUE;
		break;

	case OM_REMOVE:
		/* Remove object from list. */
		Remove ((struct Node *) _OBJECT(o));
		retval = TRUE;
		break;

	case OM_SET:
	case OM_GET:
	case OM_UPDATE:
	case OM_NOTIFY:
	case OM_ADDMEMBER:
	case OM_REMMEMBER:
		break;

	case OM_SETOBJC:
		BOOPSIEXTRA(o)->ObjectiveC = ((struct opObjc *)msg)->ObjCObject;
		retval = TRUE;
		break;
	
	case OM_GETOBJC:
		retval = (IPTR)BOOPSIEXTRA(o)->ObjectiveC;
		break;
	
	default:
		/* Ignore */
		break;

	} /* switch */

	return (retval);

	AROS_USERFUNC_EXIT

} /* rootDispatcher */
